import org.eclipse.emf.ecore.EStructuralFeature.Setting
import org.eclipse.emf.ecore.util.EcoreUtil.UsageCrossReferencer
import org.eclipse.emf.ecore.*
import org.eclipse.emf.ecore.util.*
import edu.tum.cs.cope.migration.*
import edu.tum.cs.cope.migration.execution.*


/*
 * Primitives for User Interaction
 */
 
choose = { Instance instance, List values, String message ->
	return MigratorRegistry.getInstance().getOracle().choose(instance, values, message)
}

debug = { Instance instance, String message ->
	MigratorRegistry.getInstance().getDebugger().debug(instance, message)
}

/*
 * Primitives for Metamodel Adaptation
 */

// Metamodel query

mm = { String name -> repository.metamodel.getElement(name) }
p = { String name -> repository.metamodel.getEPackage(name) }
rp = { String nsURI -> repository.metamodel.ePackages.find{p -> p.nsURI == nsURI} }
EPackage.metaClass.c = { String name -> delegate.getEClassifier(name) }
EClass.metaClass.f = { String name -> delegate.getEStructuralFeature(name) }
EClass.metaClass.o = { String name ->
	delegate.eOperations.find{ it.name == name }
}
EPackage.metaClass.getESubpackage = { String name ->
	delegate.eSubpackages.find{ it.name == name }
}

EPackage.metaClass.propertyMissing = { String name ->
	def p = delegate.getESubpackage(name)
	if(p) {
		return p
	}
	else {
		return delegate.c(name)
	}
}

EClass.metaClass.propertyMissing = { String name ->
	if(name.startsWith('_')) {
		name = name.substring(1)
	}
	def f = delegate.f(name)
	if(f) {
		return f
	}
	else {
		return delegate.o(name)
	}
}

EEnum.metaClass.propertyMissing = { String name ->
	return delegate.eLiterals.find{ it.name == name }
}
EOperation.metaClass.propertyMissing = { String name ->
	return delegate.eParameters.find{ it.name == name }
}

EModelElement.metaClass.getInverse = {EReference reference ->
	EModelElement element = delegate;
	List<EObject> inverseList = new ArrayList<EObject>();
	for(Setting setting in UsageCrossReferencer.find(element, repository.metamodel.ePackages)) {
		if(setting.eStructuralFeature == reference) {
			inverseList.add(setting.eObject)
		}
	}
	return inverseList;
}

EClass.metaClass.getESubTypes = { ->
	return delegate.getInverse(emf.EClass._eSuperTypes)
}

EClass.metaClass.getEAllSubTypes = { ->
	Collection<EClass> subTypes = new HashSet<EClass>()
	for(subType in delegate.eSubTypes) {
		subTypes.add(subType)
		subTypes.addAll(subType.eAllSubTypes)
	}
	return subTypes
}

EModelElement.metaClass.clone = { ->
	return EcoreUtil.copy(delegate)
}

// Metamodel modification

EPackage.metaClass.newEPackage = { ->
	def ePackage = EcoreFactory.eINSTANCE.createEPackage()
	delegate.eSubpackages.add(ePackage)
	ePackage
}

EPackage.metaClass.newEClass = { ->
	def eClass = EcoreFactory.eINSTANCE.createEClass()
	delegate.eClassifiers.add(eClass)
	eClass
}

EPackage.metaClass.newEEnum = { ->
	def eEnum = EcoreFactory.eINSTANCE.createEEnum()
	delegate.eClassifiers.add(eEnum)
	eEnum
}

EEnum.metaClass.newEEnumLiteral = { ->
	def eEnumLiteral = EcoreFactory.eINSTANCE.createEEnumLiteral()
	delegate.eLiterals.add(eEnumLiteral)
	eEnumLiteral
}

EClass.metaClass.newEAttribute = { ->
	def eAttribute = EcoreFactory.eINSTANCE.createEAttribute()
	delegate.eStructuralFeatures.add(eAttribute)
	eAttribute
}
EClass.metaClass.newEReference = { ->
	def eReference = EcoreFactory.eINSTANCE.createEReference()
	delegate.eStructuralFeatures.add(eReference)
	eReference
}
EClass.metaClass.newEOperation = { ->
	def eOperation = EcoreFactory.eINSTANCE.createEOperation()
	delegate.eOperations.add(eOperation)
	eOperation
}
EClass.metaClass.newEGenericType = { ->
	def eGenericType = EcoreFactory.eINSTANCE.createEGenericType()
	delegate.eGenericSuperTypes.add(eGenericType)
	eGenericType
}
EModelElement.metaClass.newEAnnotation = { ->
	def eAnnotation = EcoreFactory.eINSTANCE.createEAnnotation()
	delegate.eAnnotations.add(eAnnotation)
	eAnnotation
}
EAnnotation.metaClass.newEStringToStringMapEntry = { ->
	def entry = EcoreFactory.eINSTANCE.create(EcorePackage.eINSTANCE.getEStringToStringMapEntry())
	delegate.details.add(entry)
	entry
}
ETypedElement.metaClass.newEGenericType = { ->
	def eGenericType = EcoreFactory.eINSTANCE.createEGenericType()
	delegate.eGenericType = eGenericType
	eGenericType
}
EOperation.metaClass.newEParameter = { ->
	def eParameter = EcoreFactory.eINSTANCE.createEParameter()
	delegate.eParameters.add(eParameter)
	eParameter
}

EModelElement.metaClass.delete = { -> EcoreUtil.delete(delegate) }
EGenericType.metaClass.delete = { -> EcoreUtil.delete(delegate) }

EReference.metaClass.setEOpposite = {EReference opposite ->
	repository.model.setEOpposite(delegate, opposite)
}

newEClassImpl = {EPackage ePackage, String name, Collection<EClass> superClasses, boolean abstr ->
	def eClass = EcoreFactory.eINSTANCE.createEClass()
	ePackage.eClassifiers.add(eClass)
	eClass.name = name
	eClass.eSuperTypes.addAll(superClasses)
	eClass.'abstract' = abstr
	eClass
}

EPackage.metaClass.newEClass = {String name, Collection<EClass> superClasses, boolean abstr ->
	newEClassImpl(delegate, name, superClasses, abstr)
}

EPackage.metaClass.newEClass = {String name, Collection<EClass> superClasses ->
	newEClassImpl(delegate, name, superClasses, false)
}

EPackage.metaClass.newEClass = {String name ->
	newEClassImpl(delegate, name, [], false)
}

EPackage.metaClass.newEEnum = {String name ->
	def ePackage = delegate
	def eEnum = EcoreFactory.eINSTANCE.createEEnum()
	ePackage.eClassifiers.add(eEnum)
	eEnum.name = name
	eEnum
}

EEnum.metaClass.newEEnumLiteral = {String name ->
	def eEnum = delegate
	def eEnumLiteral = EcoreFactory.eINSTANCE.createEEnumLiteral()
	eEnum.eLiterals.add(eEnumLiteral)
	eEnumLiteral.name = name
	eEnumLiteral
}

newEReferenceImpl = {EClass eClass, String name, EClass type, int lowerBound, int upperBound, boolean containment, EReference opposite ->
	def reference = eClass.newEReference()
	reference.name = name
	reference.eType = type
	reference.lowerBound = lowerBound
	reference.upperBound = upperBound
	reference.containment = containment
	eClass.eStructuralFeatures.add(reference)
	if(opposite != null) {
		opposite.eOpposite = reference
	}
	reference
}

EClass.metaClass.newEReference = {String name, EClass type, int lowerBound, int upperBound, boolean containment, EReference opposite ->
	newEReferenceImpl(delegate, name, type, lowerBound, upperBound, containment, opposite)
}

EClass.metaClass.newEReference = {String name, EClass type, int lowerBound, int upperBound, boolean containment ->
	newEReferenceImpl(delegate, name, type, lowerBound, upperBound, containment, null)
}

EClass.metaClass.newEReference = {String name, EClass type, int lowerBound, int upperBound ->
	newEReferenceImpl(delegate, name, type, lowerBound, upperBound, false, null)
}

EClass.metaClass.newEReference = {String name, EClass type, int lowerBound ->
	newEReferenceImpl(delegate, name, type, lowerBound, 1, false, null)
}

EClass.metaClass.newEReference = {String name, EClass type ->
	newEReferenceImpl(delegate, name, type, 0, 1, false, null)
}

newEAttributeImpl = {EClass eClass, String name, EDataType type, int lowerBound, int upperBound, String defaultValue ->
	def eAttribute = eClass.newEAttribute()
	eAttribute.name = name
	eAttribute.eType = type
	eAttribute.lowerBound = lowerBound
	eAttribute.upperBound = upperBound
	eAttribute.defaultValueLiteral = defaultValue
	eAttribute	
}

EClass.metaClass.newEAttribute = {String name, EDataType type, int lowerBound, int upperBound, String defaultValue ->
	newEAttributeImpl(delegate, name, type, lowerBound, upperBound, defaultValue)
}

EClass.metaClass.newEAttribute = {String name, EDataType type, int lowerBound, int upperBound ->
	newEAttributeImpl(delegate, name, type, lowerBound, upperBound, null)
}

EClass.metaClass.newEAttribute = {String name, EDataType type, int lowerBound ->
	newEAttributeImpl(delegate, name, type, lowerBound, 1, null)
}

EClass.metaClass.newEAttribute = {String name, EDataType type ->
	newEAttributeImpl(delegate, name, type, 0, 1, null)
}

EClass.metaClass.newEOperation = {String name, EClassifier type, int lowerBound, int upperBound ->
	def eClass = delegate
	def eOperation = eClass.newEOperation()
	eOperation.name = name
	eOperation.eType = type
	eOperation.lowerBound = lowerBound
	eOperation.upperBound = upperBound
	eOperation
}

EOperation.metaClass.newEParameter = {String name, EClassifier type, int lowerBound, int upperBound ->
	def eOperation = delegate
	def eParameter = eOperation.newEParameter()
	eParameter.name = name
	eParameter.eType = type
	eParameter.lowerBound = lowerBound
	eParameter.upperBound = upperBound
	eParameter
}

EModelElement.metaClass.newEAnnotation = {String source ->
	def eModelElement = delegate
	def eAnnotation = EcoreFactory.eINSTANCE.createEAnnotation()
	eModelElement.eAnnotations.add(eAnnotation)
	eAnnotation.source = source
	eAnnotation
}

EAnnotation.metaClass.newEStringToStringMapEntry = {String key, String value ->
	def eAnnotation = delegate
	def entry = eAnnotation.newEStringToStringMapEntry()
	entry.key = key
	entry.value = value
	entry
}


/**
 * Primitives for Model Migration
 */
 
// Model query

Instance.metaClass.propertyMissing = { String name ->
	if(name.startsWith('_')) {
		name = name.substring(1)
	}
	delegate.get(name)
}

EClass.metaClass.getInstances = { -> repository.model.getExactInstances(delegate) }
EClass.metaClass.getAllInstances = { -> repository.model.getInstances(delegate) }

// Model modification

Instance.metaClass.setProperty = { String name, Object value ->
	if(name.startsWith('_')) {
		name = name.substring(1)
	}
	delegate.set(name, value)
}

EClass.metaClass.newInstance = { -> repository.model.newInstance(delegate) }

Instance.metaClass.delete = {
	repository.model.delete(delegate)
}


//copy a list of instances
Instance.metaClass.clone = { ->

	// copy the tree structure with an instance as root
	copyTree = {Instance original, Map<Instance,Instance> map ->
		def EClass eClass = original.eClass
		def Instance copi = eClass.newInstance()
		for(EReference reference in eClass.getEAllReferences()) {
			if(reference.containment) {
				if(reference.many) {
					for(Instance child in original.get(reference)) {
						copi.get(reference).add(copyTree(child, map))
					}
				}
				else {
					def Instance child = original.get(reference)
					if(child != null) {
						copi.set(reference, copyTree(child, map))
					}
				}
			}
		}
		for(EAttribute attribute in eClass.getEAllAttributes()) {
			copi.set(attribute, original.get(attribute))
		}
		map[original] = copi
		return copi
	}
	
	// copy cross references of an instance
	copyReferences = {Instance original, Map<Instance,Instance> map ->
		def EClass eClass = original.eClass
		def Instance copi = map[original]
		for(EReference reference in eClass.getEAllReferences()) {
			if(!reference.containment) {
				if(reference.many) {
					if(reference.eOpposite == null || reference.eOpposite.many) {
						for(Instance ref in original.get(reference)) {
							if(map[ref] != null) {
								ref = map[ref]
							}
							copi.get(reference).add(ref)
						}
					}
				}
				else {
					if(reference.eOpposite == null || !reference.eOpposite.containment) {
						def Instance ref = original.get(reference)
						if(ref != null) {
							if(map[ref] != null) {
								ref = map[ref]
							}
							copi.set(reference, ref)
						}
					}
				}
			}
		}
	}

	def original = delegate

	// mapping of originals to copies
	Map<Instance,Instance> map = new HashMap<Instance,Instance>()
	
	// copy tree structure
	def copy = copyTree(original, map)
	
	// copy cross references
	copyReferences(original, map)

	return copy
}